# stdlib
import base64
import secrets
from timeit import timeit
from typing import Any as TypeAny
from typing import Dict as TypeDict
from typing import Generator
from typing import List as TypeList
from typing import Union

# third party
import requests
from syft_proto.execution.v1.plan_pb2 import Plan as PlanTorchscriptPB

# syft relative
from ..core.plan import Plan
from ..core.plan.translation.torchscript.plan import PlanTorchscript
from ..federated.model_centric_fl_base import ModelCentricFLBase
from ..lib.python.list import List
from ..proto.core.plan.plan_pb2 import Plan as PlanPB
from .model_serialization import deserialize_model_params
from .model_serialization import wrap_model_params

CHUNK_SIZE = 655360  # 640KB
SPEED_MULT_FACTOR = 10
MAX_BUFFER_SIZE = 1048576 * 64  # 64 MB
MAX_SPEED_TESTS = 3


class ModelCentricFLWorker(ModelCentricFLBase):
    CYCLE_STATUS_ACCEPTED = "accepted"
    CYCLE_STATUS_REJECTED = "rejected"
    PLAN_TYPE_LIST = "list"
    PLAN_TYPE_TORCHSCRIPT = "torchscript"

    def _yield_chunk_from_request(
        self, request: requests.Response, chunk_size: int
    ) -> Generator:
        for chunk in request.iter_content(chunk_size=chunk_size):
            yield chunk

    def _read_n_request_chunks(self, chunk_generator: Generator, n: int) -> bool:
        for i in range(n):
            try:
                next(chunk_generator)
            except Exception:
                return False
        return True

    def _get_ping(self, worker_id: str, random_id: int) -> float:
        params = {"is_ping": 1, "worker_id": worker_id, "random": random_id}
        ping = (
            timeit(
                lambda: self._send_http_req("GET", "/model-centric/speed-test", params),
                number=MAX_SPEED_TESTS,
            )
            * 1000
        )  # for ms
        return ping

    def _get_upload_speed(self, worker_id: str, random_id: int) -> float:
        buffer_size = CHUNK_SIZE
        speed_history = []

        for _ in range(MAX_SPEED_TESTS):
            data_sample = b"x" * buffer_size
            params = {"worker_id": worker_id, "random": random_id}
            body = {"upload_data": data_sample}
            time_taken = timeit(
                lambda: self._send_http_req(
                    "POST", "/model-centric/speed-test", params, body
                ),
                number=1,
            )
            if time_taken < 0.5:
                buffer_size = min(buffer_size * SPEED_MULT_FACTOR, MAX_BUFFER_SIZE)
            new_speed = buffer_size / (time_taken * 1024)

            if new_speed != float("inf"):
                speed_history.append(new_speed)

        if len(speed_history) == 0:
            return float("inf")
        else:
            avg_speed = sum(speed_history) / len(speed_history)
            return avg_speed

    def _get_download_speed(self, worker_id: str, random_id: int) -> float:
        params: TypeDict[str, Union[int, str]] = {
            "worker_id": worker_id,
            "random": random_id,
        }
        speed_history = []
        with requests.get(
            self.http_url + "/model-centric/speed-test", params, stream=True
        ) as r:
            r.raise_for_status()
            buffer_size = CHUNK_SIZE
            chunk_generator = self._yield_chunk_from_request(r, CHUNK_SIZE)
            for _ in range(MAX_SPEED_TESTS):
                time_taken = timeit(
                    lambda: self._read_n_request_chunks(
                        chunk_generator, buffer_size // CHUNK_SIZE
                    ),
                    number=1,
                )
                if time_taken < 0.5:
                    buffer_size = min(buffer_size * SPEED_MULT_FACTOR, MAX_BUFFER_SIZE)
                new_speed = buffer_size / (time_taken * 1024)

                if new_speed != float("inf"):
                    speed_history.append(new_speed)

        if len(speed_history) == 0:
            return float("inf")
        else:
            avg_speed = sum(speed_history) / len(speed_history)
            return avg_speed

    def authenticate(
        self, auth_token: str, model_name: str, model_version: str
    ) -> TypeDict[str, TypeAny]:
        message = {
            "type": "model-centric/authenticate",
            "data": {
                "auth_token": auth_token,
                "model_name": model_name,
                "model_version": model_version,
            },
        }

        return self._send_msg(message)

    def cycle_request(
        self,
        worker_id: str,
        model_name: str,
        model_version: str,
        speed_info: TypeDict[str, TypeAny],
    ) -> TypeDict[str, TypeAny]:
        message = {
            "type": "model-centric/cycle-request",
            "data": {
                "worker_id": worker_id,
                "model": model_name,
                "version": model_version,
                **speed_info,
            },
        }
        return self._send_msg(message)

    def get_model(self, worker_id: str, request_key: str, model_id: int) -> TypeList:
        params_dict = {
            "worker_id": worker_id,
            "request_key": request_key,
            "model_id": model_id,
        }
        serialized_model = self._send_http_req(
            "GET", "/model-centric/get-model", params_dict
        )
        # TODO migrate to syft-core protobufs
        params: List = deserialize_model_params(serialized_model)
        return params.upcast()

    def get_plan(
        self, worker_id: str, request_key: str, plan_id: int, receive_operations_as: str
    ) -> Union[PlanTorchscript, Plan]:
        params = {
            "worker_id": worker_id,
            "request_key": request_key,
            "plan_id": plan_id,
            "receive_operations_as": receive_operations_as,
        }
        serialized_plan = self._send_http_req("GET", "/model-centric/get-plan", params)
        if receive_operations_as == ModelCentricFLWorker.PLAN_TYPE_TORCHSCRIPT:
            # TODO migrate to syft-core protobufs
            pb = PlanTorchscriptPB()
            pb.ParseFromString(serialized_plan)
            return PlanTorchscript._proto2object(pb)
        else:
            return self._unserialize(serialized_plan, PlanPB)

    def report(
        self, worker_id: str, request_key: str, diff: TypeList
    ) -> TypeDict[str, TypeAny]:
        # TODO migrate to syft-core protobufs
        diff_serialized = self._serialize(wrap_model_params(diff))
        diff_base64 = base64.b64encode(diff_serialized).decode("ascii")
        params = {
            "type": "model-centric/report",
            "data": {
                "worker_id": worker_id,
                "request_key": request_key,
                "diff": diff_base64,
            },
        }
        return self._send_msg(params)

    def get_connection_speed(self, worker_id: str) -> TypeDict[str, TypeAny]:
        random_num = secrets.randbits(128)
        ping = self._get_ping(worker_id, random_num)
        upload_speed = self._get_upload_speed(worker_id, random_num)
        download_speed = self._get_download_speed(worker_id, random_num)
        return {"ping": ping, "download": download_speed, "upload": upload_speed}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/metadata.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import location_pb2 as proto_dot_core_dot_io_dot_location__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/node/common/metadata.proto",
    package="syft.core.node.common",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n%proto/core/node/common/metadata.proto\x12\x15syft.core.node.common\x1a%proto/core/common/common_object.proto\x1a\x1cproto/core/io/location.proto"i\n\x08Metadata\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x0c\n\x04name\x18\x02 \x01(\t\x12,\n\x04node\x18\x03 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocationb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_location__pb2.DESCRIPTOR,
    ],
)


_METADATA = _descriptor.Descriptor(
    name="Metadata",
    full_name="syft.core.node.common.Metadata",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="syft.core.node.common.Metadata.id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="name",
            full_name="syft.core.node.common.Metadata.name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="node",
            full_name="syft.core.node.common.Metadata.node",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=133,
    serialized_end=238,
)

_METADATA.fields_by_name[
    "id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_METADATA.fields_by_name[
    "node"
].message_type = proto_dot_core_dot_io_dot_location__pb2._SPECIFICLOCATION
DESCRIPTOR.message_types_by_name["Metadata"] = _METADATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Metadata = _reflection.GeneratedProtocolMessageType(
    "Metadata",
    (_message.Message,),
    {
        "DESCRIPTOR": _METADATA,
        "__module__": "proto.core.node.common.metadata_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.Metadata)
    },
)
_sym_db.RegisterMessage(Metadata)


# @@protoc_insertion_point(module_scope)



# stdlib
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# syft relative
from ... import serialize
from ...proto.core.store.dataset_pb2 import Dataset as Dataset_PB
from ...util import get_fully_qualified_name
from ..common.serde.deserialize import _deserialize
from ..common.serde.serializable import Serializable
from ..common.serde.serializable import bind_protobuf
from ..common.uid import UID
from .storeable_object import StorableObject


@bind_protobuf
class Dataset(Serializable):
    """
    Dataset is a wrapper over a collection of Serializable objects.

    Arguments:
        id (UID): the id at which to store the data.
        data (List[Serializable]): A list of serializable objects.
        description (Optional[str]): An optional string that describes what you are storing. Useful
        when searching.
        tags (Optional[List[str]]): An optional list of strings that are tags used at search.
        TODO: add docs about read_permission and search_permission

    Attributes:
        id (UID): the id at which to store the data.
        data (List[Serializable]): A list of serializable objects.
        description (Optional[str]): An optional string that describes what you are storing. Useful
        when searching.
        tags (Optional[List[str]]): An optional list of strings that are tags used at search.

    """

    def __init__(
        self,
        id: UID,
        data: List[StorableObject],
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        read_permissions: Optional[dict] = None,
        search_permissions: Optional[dict] = None,
    ):
        self.id = id
        self.data = data
        self._description: str = description if description else ""
        self._tags: List[str] = tags if tags else []

        # the dict key of "verify key" objects corresponding to people
        # the value is the original request_id to allow lookup later
        # who are allowed to call .get() and download this object.
        self.read_permissions = read_permissions if read_permissions else {}

        # the dict key of "verify key" objects corresponding to people
        # the value is the original request_id to allow lookup later
        # who are allowed to know that the tensor exists (via search or other means)
        self.search_permissions: dict = search_permissions if search_permissions else {}

    @property
    def tags(self) -> Optional[List[str]]:
        return self._tags

    @tags.setter
    def tags(self, value: Optional[List[str]]) -> None:
        self._tags = value if value else []

    @property
    def description(self) -> Optional[str]:
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description if description else ""

    @property
    def class_name(self) -> str:
        return str(self.__class__.__name__)

    def __contains__(self, _id: UID) -> bool:
        return _id in [el.id for el in self.data]

    def keys(self) -> List[UID]:
        return [el.id for el in self.data]

    def __getitem__(self, _id: UID) -> List[StorableObject]:
        return [el for el in self.data if el.id == _id]

    def __delitem__(self, _id: UID) -> None:
        self.data = [el for el in self.data if el.id != _id]

    def _object2proto(self) -> Dataset_PB:
        proto = Dataset_PB()

        # Step 1: Serialize the id to protobuf and copy into protobuf
        id = serialize(self.id)
        proto.id.CopyFrom(id)

        # Step 2: Save the type of wrapper to use to deserialize
        proto.obj_type = get_fully_qualified_name(obj=self)

        # Step 3: Serialize data to protobuf and pack into proto
        if hasattr(self, "data"):
            if self.data is not None:
                for _d in self.data:
                    proto_storable = _d._object2proto()
                    proto.data.append(proto_storable)

        if hasattr(self, "description"):
            # Step 4: save the description into proto
            proto.description = self.description

        # QUESTION: Which one do we want, self.data.tags or self.tags or both???
        if hasattr(self, "tags"):
            # Step 5: save tags into proto if they exist
            if self.tags is not None:
                for tag in self.tags:
                    proto.tags.append(tag)

        # Step 6: save read permissions
        if len(self.read_permissions.keys()) > 0:
            permission_data = sy.lib.python.Dict()
            for k, v in self.read_permissions.items():
                permission_data[k] = v
            proto.read_permissions = serialize(permission_data, to_bytes=True)

        # Step 7: save search permissions
        if len(self.search_permissions.keys()) > 0:
            permission_data = sy.lib.python.Dict()
            for k, v in self.search_permissions.items():
                permission_data[k] = v
            proto.search_permissions = serialize(permission_data, to_bytes=True)

        return proto

    @staticmethod
    def _proto2object(proto: Dataset_PB) -> Serializable:

        # Step 1: deserialize the ID
        id = _deserialize(blob=proto.id)

        if not isinstance(id, UID):
            raise ValueError("TODO")

        # Step 2: Deserialize data from protobuf
        data = list(proto.data) if proto.data else []
        data = [StorableObject._proto2object(proto=d) for d in data]

        # Step 3: get the description from proto
        description = proto.description if proto.description else ""

        # Step 4: get the tags from proto of they exist
        tags = list(proto.tags) if proto.tags else []

        result = Dataset(id=id, data=data, description=description, tags=tags)
        return result

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """
        return Dataset_PB



"""In this test suite, we evaluate the ObjectWithID class. For more info
on the ObjectWithID class and its purpose, please see the documentation
in the class itself.

Table of Contents:
    - INITIALIZATION: tests for various ways ObjectWithID can/can't be initialized
    - CLASS METHODS: tests for the use of ObjectWithID's class methods
    - SERDE: test for serialization and deserialization of ObjectWithID.
    - CHILDREN: test that subclasses of ObjectWithID fulfil standards

"""

# stdlib
import uuid

# third party
import pytest

# syft absolute
import syft as sy
from syft.core.common import ObjectWithID
from syft.core.common import UID
from syft.util import get_subclasses

# --------------------- INITIALIZATION ---------------------


def test_basic_init() -> None:
    """Test that creating ObjectWithID() does in fact create
    an object with an id."""

    obj = ObjectWithID()
    assert isinstance(obj.id, UID)


def test_immutability_of_id() -> None:
    """We shouldn't allow people to modify the id of an
    ObjectWithID because this can create all sorts of errors.

    Put the other way around - blocking people from modifying
    the ID of an object means we can build a codebase which more
    firmly relies on the id being truthful. It also will avoid
    people initialising objects in weird ways (setting ids later).
    """
    obj = ObjectWithID()

    with pytest.raises(AttributeError):

        # TODO: filter on this error to only include errors
        #  with the string "Can't set attribute"

        obj.id = ""


# --------------------- CLASS METHODS ---------------------


def test_compare() -> None:
    """While uses of this feature should be quite rare, we
    should be able to check whether two objects are the same
    based on their IDs being the same by default. Note that
    subclasses will undoubtedly modify this behavior with other
    __eq__ methods."""

    obj = ObjectWithID()
    obj2 = ObjectWithID()

    assert obj != obj2

    obj._id = obj2.id

    assert obj == obj2


def test_to_string() -> None:
    """Tests that UID generates an intuitive string."""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = ObjectWithID(id=uid)
    assert str(obj) == "<ObjectWithID: fb1bb0675bb74c49becee700ab0a1514>"
    assert obj.__repr__() == "<ObjectWithID: fb1bb0675bb74c49becee700ab0a1514>"


# --------------------- SERDE ---------------------


def test_object_with_id_default_serialization() -> None:
    """Tests that default ObjectWithID serialization works as expected - to Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = ObjectWithID(id=uid)

    blob = sy.serialize(obj, to_proto=True)

    assert sy.serialize(obj) == blob


def test_object_with_id_default_deserialization() -> None:
    """Tests that default ObjectWithID deserialization works as expected - from Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = ObjectWithID(id=uid)

    blob = ObjectWithID.get_protobuf_schema()(id=sy.serialize(uid))

    obj2 = sy.deserialize(blob=blob)
    assert obj == obj2


def test_object_with_id_proto_serialization() -> None:
    """Tests that default ObjectWithID serialization works as expected - to Protobuf"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = ObjectWithID(id=uid)

    blob = ObjectWithID.get_protobuf_schema()(id=sy.serialize(uid))

    assert sy.serialize(obj, to_proto=True) == blob
    assert sy.serialize(obj, to_proto=True) == blob
    assert sy.serialize(obj, to_proto=True) == blob


def test_object_with_id_proto_deserialization() -> None:
    """Tests that default UID deserialization works as expected"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = ObjectWithID(id=uid)

    blob = ObjectWithID.get_protobuf_schema()(id=sy.serialize(uid))

    obj2 = sy.deserialize(blob=blob, from_proto=True)
    assert obj == obj2


def test_object_with_id_binary_serialization() -> None:
    """Tests that binary ObjectWithID serializes as expected"""

    uid = UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    obj = ObjectWithID(id=uid)

    blob = (
        b"\n$syft.core.common.object.ObjectWithID\x12\x14\n\x12\n\x10\xfb\x1b\xb0"
        + b"g[\xb7LI\xbe\xce\xe7\x00\xab\n\x15\x14"
    )

    assert sy.serialize(obj, to_bytes=True) == blob
    assert sy.serialize(obj, to_bytes=True) == blob
    assert sy.serialize(obj, to_bytes=True) == blob


def test_object_with_id_binary_deserialization() -> None:
    """Test that binary ObjectWithID deserialization works as expected"""

    blob = (
        b"\n$syft.core.common.object.ObjectWithID\x12\x14\n\x12\n\x10\xfb\x1b\xb0"
        + b"g[\xb7LI\xbe\xce\xe7\x00\xab\n\x15\x14"
    )
    obj = sy.deserialize(blob=blob, from_bytes=True)
    assert obj == ObjectWithID(
        id=UID(value=uuid.UUID(int=333779996850170035686993356951732753684))
    )


# ----------------------- CHILDREN -----------------------


def test_subclasses_have_names() -> None:
    """Ensure that all known subclassses of ObjectWithID have
    a __name__ parameter. I'm not sure why but occasionally
    I came across objects without them"""

    subclasses = get_subclasses(obj_type=ObjectWithID)

    for sc in subclasses:
        assert hasattr(sc, "__name__")



"""
Tests copied from cpython test suite:
https://github.com/python/cpython/blob/3.9/Lib/test/test_float.py
"""

# stdlib
from math import copysign
from math import isinf
from math import isnan
from math import ldexp
import operator
import random
import sys
import time

# third party
import pytest

# syft absolute
from syft.lib.python.float import Float

VALID_UNDERSCORE_LITERALS = [
    "0_0_0",
    "4_2",
    "1_0000_0000",
    "0b1001_0100",
    "0xffff_ffff",
    "0o5_7_7",
    "1_00_00.5",
    "1_00_00.5e5",
    "1_00_00e5_1",
    "1e1_0",
    ".1_4",
    ".1_4e1",
    "0b_0",
    "0x_f",
    "0o_5",
    "1_00_00j",
    "1_00_00.5j",
    "1_00_00e5_1j",
    ".1_4j",
    "(1_2.5+3_3j)",
    "(.5_6j)",
]

INVALID_UNDERSCORE_LITERALS = [
    # Trailing underscores:
    "0_",
    "42_",
    "1.4j_",
    "0x_",
    "0b1_",
    "0xf_",
    "0o5_",
    "0 if 1_Else 1",
    # Underscores in the base selector:
    "0_b0",
    "0_xf",
    "0_o5",
    # Old-style octal, still disallowed:
    "0_7",
    "09_99",
    # Multiple consecutive underscores:
    "4_______2",
    "0.1__4",
    "0.1__4j",
    "0b1001__0100",
    "0xffff__ffff",
    "0x___",
    "0o5__77",
    "1e1__0",
    "1e1__0j",
    # Underscore right before a dot:
    "1_.4",
    "1_.4j",
    # Underscore right after a dot:
    "1._4",
    "1._4j",
    "._5",
    "._5j",
    # Underscore right after a sign:
    "1.0e+_1",
    "1.0e+_1j",
    # Underscore right before j:
    "1.4_j",
    "1.4e5_j",
    # Underscore right before e:
    "1_e1",
    "1.4_e1",
    "1.4_e1j",
    # Underscore right after e:
    "1e_1",
    "1.4e_1",
    "1.4e_1j",
    # Complex cases with parens:
    "(1+1.5_j_)",
    "(1+1.5_j)",
]

INF = Float("inf")
NAN = Float("nan")

have_getformat = hasattr(Float, "__getformat__")


class FloatSubclass(Float):
    pass


class OtherFloatSubclass(Float):
    pass


class F:
    def __float__(self):
        return OtherFloatSubclass(42.0)


def test_Float():
    assert Float(3.14) == 3.14
    assert Float(314) == 314.0
    assert Float("  3.14  ") == 3.14
    with pytest.raises(ValueError):
        Float("  0x3.1  ")
    with pytest.raises(ValueError):
        Float("  -0x3.p-1  ")
    with pytest.raises(ValueError):
        Float("  +0x3.p-1  ")
    with pytest.raises(ValueError):
        Float("++3.14")
    with pytest.raises(ValueError):
        Float("+-3.14")
    with pytest.raises(ValueError):
        Float("-+3.14")
    with pytest.raises(ValueError):
        Float("--3.14")
    with pytest.raises(ValueError):
        Float(".nan")
    with pytest.raises(ValueError):
        Float("+.inf")
    with pytest.raises(ValueError):
        Float(".")
    with pytest.raises(ValueError):
        Float("-.")
    with pytest.raises(TypeError):
        Float({})

    with pytest.raises(TypeError):
        Float({})
    # assertRaisesRegex(TypeError, "not 'dict'", Float, {})
    # Lone surrogate
    with pytest.raises(ValueError):
        Float("\uD8F0")

    # check that we don't accept alternate exponent markers
    with pytest.raises(ValueError):
        Float("-1.7d29")

    with pytest.raises(ValueError):
        Float("3D-14")

    assert Float("  \u0663.\u0661\u0664  ") == 3.14
    assert Float("\N{EM SPACE}3.14\N{EN SPACE}") == 3.14

    # extra long strings should not be a problem
    Float(b"." + b"1" * 1000)
    Float("." + "1" * 1000)

    with pytest.raises(ValueError):
        Float("\u3053\u3093\u306b\u3061\u306f")


def test_underscores():
    for lit in VALID_UNDERSCORE_LITERALS:
        if not any(ch in lit for ch in "jJxXoObB"):
            assert Float(lit) == eval(lit)
            # TODO this should work!!!
            assert Float(lit) == Float(lit.replace("_", ""))
    for lit in INVALID_UNDERSCORE_LITERALS:
        if lit in ("0_7", "09_99"):  # octals are not recognized here
            continue
        if not any(ch in lit for ch in "jJxXoObB"):
            with pytest.raises(ValueError):
                Float(lit)
    # Additional test cases; nan and inf are never valid as literals,
    # only in the Float() constructor, but we don't allow underscores
    # in or around them.
    with pytest.raises(ValueError):
        Float("_NaN")
    with pytest.raises(ValueError):
        Float("Na_N")
    with pytest.raises(ValueError):
        Float("IN_F")
    with pytest.raises(ValueError):
        Float("-_INF")
    with pytest.raises(ValueError):
        Float("-INF_")
    # Check that we handle bytes values correctly.
    with pytest.raises(ValueError):
        Float(b"0_.\xff9")


def test_non_numeric_input_types():
    # Test possible non-numeric types for the argument x, including
    # subclasses of the explicitly documented accepted types.
    class CustomStr(str):
        pass

    class CustomBytes(bytes):
        pass

    class CustomByteArray(bytearray):
        pass

    factories = [
        bytes,
        bytearray,
        lambda b: CustomStr(b.decode()),
        CustomBytes,
        CustomByteArray,
        memoryview,
    ]
    try:
        # stdlib
        from array import array
    except ImportError:
        pass
    else:
        factories.append(lambda b: array("B", b))

    for f in factories:
        x = f(b" 3.14  ")
        assert Float(x) == 3.14
        with pytest.raises(ValueError):
            Float(f(b"A" * 0x10))


def test_Float_memoryview():
    assert Float(memoryview(b"12.3")[1:4]) == 2.3
    assert Float(memoryview(b"12.3\x00")[1:4]) == 2.3
    assert Float(memoryview(b"12.3 ")[1:4]) == 2.3
    assert Float(memoryview(b"12.3A")[1:4]) == 2.3
    assert Float(memoryview(b"12.34")[1:4]) == 2.3


def test_error_message():
    def check(s):
        with pytest.raises(ValueError):
            Float(s)

    check("\xbd")
    check("123\xbd")
    check("  123 456  ")
    check(b"  123 456  ")

    # non-ascii digits (error came from non-digit '!')
    check("\u0663\u0661\u0664!")
    # embedded NUL
    check("123\x00")
    check("123\x00 245")
    check("123\x00245")
    # byte string with embedded NUL
    check(b"123\x00")
    # non-UTF-8 byte string
    check(b"123\xa0")


def test_Float_with_comma():
    # set locale to something that doesn't use '.' for the decimal point    assert Float("  3.14  ") == 3.14
    # Float must not accept the locale specific decimal point but
    # it still has to accept the normal python syntax
    assert Float("+3.14  ") == 3.14
    assert Float("-3.14  ") == -3.14
    assert Float(".14  ") == 0.14
    assert Float("3.  ") == 3.0
    assert Float("3.e3  ") == 3000.0
    assert Float("3.2e3  ") == 3200.0
    assert Float("2.5e-1  ") == 0.25
    assert Float("5e-1") == 0.5
    with pytest.raises(ValueError):
        Float("  3,14  ")
    with pytest.raises(ValueError):
        Float("  +3,14  ")
    with pytest.raises(ValueError):
        Float("  -3,14  ")
    with pytest.raises(ValueError):
        Float("  0x3.1  ")
    with pytest.raises(ValueError):
        Float("  -0x3.p-1  ")
    with pytest.raises(ValueError):
        Float("  +0x3.p-1  ")
    assert Float("  25.e-1  ") == 2.5


def test_Floatconversion():
    # Make sure that calls to __Float__() work properly
    class Foo1(object):
        def __float__(self):
            return 42.0

    class Foo2(Float):
        def __float__(self):
            return 42.0

    class Foo3(Float):
        def __new__(cls, value=0.0):
            return Float.__new__(cls, 2 * value)

        def __float__(self):
            return self

    class Foo4(Float):
        def __float__(self):
            return 42

    # Issue 5759: __Float__ not called on str subclasses (though it is on
    # unicode subclasses).
    class FooStr(str):
        def __float__(self):
            return Float(str(self)) + 1

    assert Float(Foo1()) == 42.0
    assert Float(Foo2()) == 42.0

    with pytest.raises(TypeError):
        Float(Foo4(42))
    assert Float(FooStr("8")) == 9.0

    class Foo5:
        def __Float__(self):
            return ""

    with pytest.raises(TypeError):
        time.sleep(Foo5())

    # using __index__ in init was added in python 3.8
    # https://github.com/python/cpython/commit/bdbad71b9def0b86433de12cecca022eee91bd9f

    if sys.version_info >= (3, 8):

        class MyIndex:
            def __init__(self, value):
                self.value = value

            def __index__(self):
                return self.value

        assert Float(MyIndex(42)) == 42.0
        with pytest.raises(OverflowError):
            Float(MyIndex(2 ** 2000))

    class MyInt:
        def __int__(self):
            return 42

    with pytest.raises(TypeError):
        Float(MyInt())


def test_keyword_args():
    with pytest.raises(TypeError):
        Float(x="3.14")


def test_is_integer():
    assert not Float(1.1).is_integer()
    assert Float(1.0).is_integer()
    assert not Float("nan").is_integer()
    assert not Float("inf").is_integer()


def test_Floatasratio():
    for f, ratio in [
        (0.875, (7, 8)),
        (-0.875, (-7, 8)),
        (0.0, (0, 1)),
        (11.5, (23, 2)),
    ]:
        assert f.as_integer_ratio() == ratio

    for i in range(10000):
        f = random.random()
        f *= 10 ** random.randint(-100, 100)
        n, d = f.as_integer_ratio()
        assert Float(n).__truediv__(d) == f

    with pytest.raises(OverflowError):
        Float("inf").as_integer_ratio()
    with pytest.raises(OverflowError):
        Float("-inf").as_integer_ratio()
    with pytest.raises(ValueError):
        Float("nan").as_integer_ratio()


def test_Float_containment():
    Floats = (INF, -INF, 0.0, 1.0, NAN)
    for f in Floats:
        assert f in [f]
        assert f in (f,)
        assert f in {f}
        assert f in {f: None}
        assert [f].count(f), 1 == f"[].count('{f!r}') != 1"
        assert f in Floats

    for f in Floats:
        # nonidentical containers, same type, same contents
        assert [f] == [f]
        assert (f,) == (f,)
        assert {f} == {f}
        assert {f: None} == {f: None}
        # identical containers
        l, t, s, d = [f], (f,), {f}, {f: None}
        assert l == l  # noqa: E741
        assert t == t  # noqa: E741
        assert s == s  # noqa: E741
        assert d == d  # noqa: E741


def test_Float_mod():
    # Check behaviour of % operator for IEEE 754 special cases.
    # In particular, check signs of zeros.
    mod = operator.mod

    assert mod(Float(-1.0), Float(1.0)) == 0.0
    assert mod(Float(-1e-100), Float(1.0)) == 1.0
    assert mod(Float(-0.0), Float(1.0)) == 0.0
    assert mod(Float(0.0), Float(1.0)) == 0.0
    assert mod(Float(1e-100), Float(1.0)) == 1e-100
    assert mod(Float(1.0), Float(1.0)) == 0.0

    assert mod(Float(-1.0), Float(-1.0)) == -0.0
    assert mod(Float(-1e-100), Float(-1.0)) == -1e-100
    assert mod(Float(-0.0), Float(-1.0)) == -0.0
    assert mod(Float(0.0), Float(-1.0)) == -0.0
    assert mod(Float(1e-100), Float(-1.0)) == -1.0
    assert mod(Float(1.0), Float(-1.0)) == -0.0


def test_Float_pow():
    # test builtin pow and ** operator for IEEE 754 special cases.
    # Special cases taken from section F.9.4.4 of the C99 specification

    for pow_op in pow, operator.pow:
        # x**NAN is NAN for any x except 1
        assert isnan(pow_op(-INF, NAN))
        assert isnan(pow_op(Float(Float(-2.0)), NAN))
        assert isnan(pow_op(Float(-1.0), NAN))
        assert isnan(pow_op(Float(-0.5), NAN))
        assert isnan(pow_op(Float(0.5), NAN))
        assert isnan(pow_op(Float(2.0), NAN))
        assert isnan(pow_op(INF, NAN))
        assert isnan(pow_op(NAN, NAN))

        # NAN**y is NAN for any y except +-0
        assert isnan(pow_op(NAN, -INF))
        assert isnan(pow_op(NAN, Float(-2.0)))
        assert isnan(pow_op(NAN, Float(-1.0)))
        assert isnan(pow_op(NAN, Float(-0.5)))
        assert isnan(pow_op(NAN, Float(0.5)))
        assert isnan(pow_op(NAN, Float(1.0)))
        assert isnan(pow_op(NAN, Float(2.0)))
        assert isnan(pow_op(NAN, INF))

        # (+-0)**y raises ZeroDivisionError for y a negative odd integer
        with pytest.raises(ZeroDivisionError):
            pow_op(Float(-0.0), Float(-1.0))

        with pytest.raises(ZeroDivisionError):
            pow_op(Float(0.0), Float(-1.0))

        # (+-0)**y raises ZeroDivisionError for y finite and negative
        # but not an odd integer
        with pytest.raises(ZeroDivisionError):
            pow_op(Float(-0.0), Float(-2.0))
        with pytest.raises(ZeroDivisionError):
            pow_op(Float(-0.0), Float(-0.5))
        with pytest.raises(ZeroDivisionError):
            pow_op(Float(0.0), Float(-2.0))
        with pytest.raises(ZeroDivisionError):
            pow_op(Float(0.0), Float(-0.5))

        # (+-0)**y is +-0 for y a positive odd integer
        assert pow_op(Float(-0.0), Float(1.0)) == -0.0
        assert pow_op(Float(0.0), Float(1.0)) == 0.0
        # (+-0)**y is 0 for y finite and positive but not an odd integer
        assert pow_op(Float(-0.0), Float(0.5)) == 0.0
        assert pow_op(Float(-0.0), Float(2.0)) == 0.0
        assert pow_op(Float(0.0), Float(0.5)) == 0.0
        assert pow_op(Float(0.0), Float(2.0)) == 0.0

        # (-1)**+-inf is 1
        assert pow_op(Float(-1.0), -INF) == 1.0

        # 1**y is 1 for any y, even if y is an infinity or nan
        assert pow_op(Float(1.0), -INF) == 1.0
        assert pow_op(Float(1.0), Float(-2.0)) == 1.0
        assert pow_op(Float(1.0), Float(-1.0)) == 1.0
        assert pow_op(Float(1.0), Float(-0.5)) == 1.0
        assert pow_op(Float(1.0), Float(-0.0)) == 1.0
        assert pow_op(Float(1.0), Float(0.0)) == 1.0
        assert pow_op(Float(1.0), Float(0.5)) == 1.0
        assert pow_op(Float(1.0), Float(1.0)) == 1.0
        assert pow_op(Float(1.0), Float(2.0)) == 1.0
        assert pow_op(Float(1.0), INF) == 1.0
        assert pow_op(Float(1.0), NAN) == 1.0

        # x**+-0 is 1 for any x, even if x is a zero, infinity, or nan
        assert pow_op(-INF, Float(0.0)) == 1.0
        assert pow_op(Float(-2.0), Float(0.0)) == 1.0
        assert pow_op(Float(-1.0), Float(0.0)) == 1.0
        assert pow_op(Float(-0.5), Float(0.0)) == 1.0
        assert pow_op(Float(-0.0), Float(0.0)) == 1.0
        assert pow_op(Float(0.0), Float(0.0)) == 1.0
        assert pow_op(Float(0.5), Float(0.0)) == 1.0
        assert pow_op(Float(1.0), Float(0.0)) == 1.0
        assert pow_op(Float(2.0), Float(0.0)) == 1.0
        assert pow_op(INF, Float(0.0)) == 1.0
        assert pow_op(NAN, Float(0.0)) == 1.0
        assert pow_op(-INF, Float(-0.0)) == 1.0
        assert pow_op(Float(-2.0), Float(-0.0)) == 1.0
        assert pow_op(Float(-1.0), Float(-0.0)) == 1.0
        assert pow_op(Float(-0.5), Float(-0.0)) == 1.0
        assert pow_op(Float(-0.0), Float(-0.0)) == 1.0
        assert pow_op(Float(0.0), Float(-0.0)) == 1.0
        assert pow_op(Float(0.5), Float(-0.0)) == 1.0
        assert pow_op(Float(1.0), Float(-0.0)) == 1.0
        assert pow_op(Float(2.0), Float(-0.0)) == 1.0
        assert pow_op(INF, Float(-0.0)) == 1.0
        assert pow_op(NAN, Float(-0.0)) == 1.0

        # x**y defers to complex pow for finite negative x and
        # non-integral y.
        assert isinstance(pow_op(Float(-2.0), Float(-0.5)), complex)
        assert isinstance(pow_op(Float(-2.0), Float(0.5)), complex)
        assert isinstance(pow_op(Float(-1.0), Float(-0.5)), complex)
        assert isinstance(pow_op(-1.0, Float(0.5)), complex)
        assert isinstance(pow_op(Float(-0.5), Float(-0.5)), complex)
        assert isinstance(pow_op(Float(-0.5), Float(0.5)), complex)

        # x**-INF is INF for abs(x) < 1
        assert pow_op(Float(-0.5), -INF) == INF
        assert pow_op(Float(-0.0), -INF) == INF
        assert pow_op(Float(0.0), -INF) == INF
        assert pow_op(Float(0.5), -INF) == INF

        # x**-INF is 0 for abs(x) > 1
        assert pow_op(-INF, -INF) == 0.0
        assert pow_op(Float(-2.0), -INF) == 0.0
        assert pow_op(Float(2.0), -INF) == 0.0
        assert pow_op(INF, -INF) == 0.0

        # x**INF is 0 for abs(x) < 1
        assert pow_op(Float(-0.5), INF) == 0.0
        assert pow_op(Float(-0.0), INF) == 0.0
        assert pow_op(Float(0.0), INF) == 0.0
        assert pow_op(Float(0.5), INF) == 0.0

        # x**INF is INF for abs(x) > 1
        assert pow_op(-INF, INF) == INF
        assert pow_op(Float(-2.0), INF) == INF
        assert pow_op(Float(2.0), INF) == INF
        assert pow_op(INF, INF) == INF

        # (-INF)**y is -0.0 for y a negative odd integer
        assert pow_op(-INF, Float(-1.0)) == -0.0

        # (-INF)**y is 0.0 for y negative but not an odd integer
        assert pow_op(-INF, Float(-0.5)) == 0.0
        assert pow_op(-INF, Float(-2.0)) == 0.0

        # (-INF)**y is -INF for y a positive odd integer
        assert pow_op(-INF, Float(1.0)) == -INF

        # (-INF)**y is INF for y positive but not an odd integer
        assert pow_op(-INF, Float(0.5)) == INF
        assert pow_op(-INF, Float(2.0)) == INF

        # INF**y is INF for y positive
        assert pow_op(INF, Float(0.5)) == INF
        assert pow_op(INF, Float(1.0)) == INF
        assert pow_op(INF, Float(2.0)) == INF

        # INF**y is 0.0 for y negative
        assert pow_op(INF, Float(-2.0)) == 0.0
        assert pow_op(INF, Float(-1.0)) == 0.0
        assert pow_op(INF, Float(-0.5)) == 0.0

        # basic checks not covered by the special cases above
        assert pow_op(Float(-2.0), Float(-2.0)) == 0.25
        assert pow_op(Float(-2.0), Float(-1.0)) == Float(-0.5)
        assert pow_op(Float(-2.0), Float(-0.0)) == Float(1.0)
        assert pow_op(Float(-2.0), Float(0.0)) == Float(1.0)
        assert pow_op(Float(-2.0), Float(1.0)) == Float(-2.0)
        assert pow_op(Float(-2.0), Float(2.0)) == Float(4.0)
        assert pow_op(Float(-1.0), Float(-2.0)) == Float(1.0)
        assert pow_op(Float(-1.0), Float(-1.0)) == Float(-1.0)
        assert pow_op(Float(-1.0), Float(-0.0)) == Float(1.0)
        assert pow_op(Float(-1.0), Float(0.0)) == Float(1.0)
        assert pow_op(Float(-1.0), Float(1.0)) == Float(-1.0)
        assert pow_op(Float(-1.0), Float(2.0)) == Float(1.0)
        assert pow_op(Float(2.0), Float(-2.0)) == Float(0.25)
        assert pow_op(Float(2.0), Float(-1.0)) == Float(0.5)
        assert pow_op(Float(2.0), Float(-0.0)) == Float(1.0)
        assert pow_op(Float(2.0), Float(0.0)) == Float(1.0)
        assert pow_op(Float(2.0), Float(1.0)) == Float(2.0)
        assert pow_op(Float(2.0), Float(2.0)) == Float(4.0)

        # 1 ** large and -1 ** large; some libms apparently
        # have problems with these
        assert pow_op(Float(1.0), -1e100) == Float(1.0)
        assert pow_op(Float(1.0), 1e100) == Float(1.0)
        assert pow_op(Float(-1.0), -1e100) == Float(1.0)
        assert pow_op(Float(-1.0), 1e100) == Float(1.0)

        # check sign for results that underflow to 0
        assert pow_op(Float(-2.0), Float(-2000.0)) == Float(0.0)
        assert isinstance(pow_op(Float(-2.0), Float(-2000.5)), complex)
        assert pow_op(Float(-2.0), Float(-2001.0)) == Float(-0.0)
        assert pow_op(Float(2.0), Float(-2000.0)) == Float(0.0)
        assert pow_op(Float(2.0), Float(-2000.5)) == Float(0.0)
        assert pow_op(Float(2.0), Float(-2001.0)) == Float(0.0)
        assert pow_op(Float(-0.5), Float(2000.0)) == Float(0.0)
        assert isinstance(pow_op(Float(-0.5), Float(2000.5)), complex)
        assert pow_op(Float(-0.5), Float(2001.0)) == Float(-0.0)
        assert pow_op(Float(0.5), Float(2000.0)) == Float(0.0)
        assert pow_op(Float(0.5), Float(2000.5)) == Float(0.0)
        assert pow_op(Float(0.5), Float(2001.0)) == Float(0.0)


def test_short_repr():
    # test short Float repr introduced in Python 3.1.  One aspect
    # of this repr is that we get some degree of str -> Float ->
    # str roundtripping.  In particular, for any numeric string
    # containing 15 or fewer significant digits, those exact same
    # digits (modulo trailing zeros) should appear in the output.
    # No more repr(0.03) -> "0.029999999999999999"!

    test_strings = [
        # output always includes *either* a decimal point and at
        # least one digit after that point, or an exponent.
        "0.0",
        "1.0",
        "0.01",
        "0.02",
        "0.03",
        "0.04",
        "0.05",
        "1.23456789",
        "10.0",
        "100.0",
        # values >= 1e16 get an exponent...
        "1000000000000000.0",
        "9999999999999990.0",
        "1e+16",
        "1e+17",
        # ... and so do values < 1e-4
        "0.001",
        "0.001001",
        "0.00010000000000001",
        "0.0001",
        "9.999999999999e-05",
        "1e-05",
        # values designed to provoke failure if the FPU rounding
        # precision isn't set correctly
        "8.72293771110361e+25",
        "7.47005307342313e+26",
        "2.86438000439698e+28",
        "8.89142905246179e+28",
        "3.08578087079232e+35",
    ]

    for s in test_strings:
        negs = "-" + s
        assert s == repr(Float(s))
        assert negs == repr(Float(negs))
        # Since Python 3.2, repr and str are identical
        assert repr(Float(s)) == str(Float(s))
        assert repr(Float(negs)) == str(Float(negs))


def test_inf_nan():
    with pytest.raises(OverflowError):
        round(INF)
    with pytest.raises(OverflowError):
        round(-INF)
    with pytest.raises(ValueError):
        round(NAN)
    with pytest.raises(TypeError):
        round(INF, 0.0)
    with pytest.raises(TypeError):
        round(-INF, 1.0)
    with pytest.raises(TypeError):
        round(NAN("ceci n'est pas un integer"))
    with pytest.raises(TypeError):
        round(-0.0, 1j)


def test_large_n():
    for n in [324, 325, 400, 2 ** 31 - 1, 2 ** 31, 2 ** 32, 2 ** 100]:
        assert round(123.456, n) == 123.456
        assert round(-123.456, n) == -123.456
        assert round(1e300, n) == 1e300
        assert round(1e-320, n) == 1e-320
    assert round(1e150, 300) == 1e150
    assert round(1e300, 307) == 1e300
    assert round(-3.1415, 308) == -3.1415
    assert round(1e150, 309) == 1e150
    assert round(1.4e-315, 315) == 1e-315


def test_small_n():
    for n in [-308, -309, -400, 1 - 2 ** 31, -(2 ** 31), -(2 ** 31) - 1, -(2 ** 100)]:
        assert round(123.456, n) == 0.0
        assert round(-123.456, n) == -0.0
        assert round(1e300, n) == 0.0
        assert round(1e-320, n) == 0.0


def test_overflow():
    with pytest.raises(OverflowError):
        round(1.6e308, -308)
    with pytest.raises(OverflowError):
        round(-1.7e308, -308)


def test_previous_round_bugs():
    # particular cases that have occurred in bug reports
    assert round(562949953421312.5, 1) == 562949953421312.5
    assert round(56294995342131.5, 3) == 56294995342131.5
    # round-half-even
    assert round(Float(25.0), -1) == Float(20.0)
    assert round(Float(35.0), -1) == Float(40.0)
    assert round(Float(45.0), -1) == Float(40.0)
    assert round(Float(55.0), -1) == Float(60.0)
    assert round(Float(65.0), -1) == Float(60.0)
    assert round(Float(75.0), -1) == Float(80.0)
    assert round(Float(85.0), -1) == Float(80.0)
    assert round(Float(95.0), -1) == Float(100.0)


@pytest.mark.slow
def test_matches_Float_format():
    # round should give the same results as Float formatting
    for i in range(500):
        x = i / 1000.0
        assert Float(format(x, ".0f")) == round(x, 0)
        assert Float(format(x, ".1f")) == round(x, 1)
        assert Float(format(x, ".2f")) == round(x, 2)
        assert Float(format(x, ".3f")) == round(x, 3)

    for i in range(5, 5000, 10):
        x = i / 1000.0
        assert Float(format(x, ".0f")) == round(x, 0)
        assert Float(format(x, ".1f")) == round(x, 1)
        assert Float(format(x, ".2f")) == round(x, 2)
        assert Float(format(x, ".3f")) == round(x, 3)

    for i in range(500):
        x = random.random()
        assert Float(format(x, ".0f")) == round(x, 0)
        assert Float(format(x, ".1f")) == round(x, 1)
        assert Float(format(x, ".2f")) == round(x, 2)
        assert Float(format(x, ".3f")) == round(x, 3)


def test_None_ndigits():
    for x in round(1.23), round(1.23, None), round(1.23, ndigits=None):
        assert x == 1
        assert isinstance(x, int)
    for x in round(1.78), round(1.78, None), round(1.78, ndigits=None):
        assert x == 2
        assert isinstance(x, int)


# Beginning with Python 2.6 Float has cross platform compatible
# ways to create and represent inf and nan


def test_inf_from_str():
    assert isinf(Float("inf"))
    assert isinf(Float("+inf"))
    assert isinf(Float("-inf"))
    assert isinf(Float("infinity"))
    assert isinf(Float("+infinity"))
    assert isinf(Float("-infinity"))

    assert repr(Float("inf")) == "inf"
    assert repr(Float("+inf")) == "inf"
    assert repr(Float("-inf")) == "-inf"
    assert repr(Float("infinity")) == "inf"
    assert repr(Float("+infinity")) == "inf"
    assert repr(Float("-infinity")) == "-inf"

    assert repr(Float("INF")) == "inf"
    assert repr(Float("+Inf")) == "inf"
    assert repr(Float("-iNF")) == "-inf"
    assert repr(Float("Infinity")) == "inf"
    assert repr(Float("+iNfInItY")) == "inf"
    assert repr(Float("-INFINITY")) == "-inf"

    assert str(Float("inf")) == "inf"
    assert str(Float("+inf")) == "inf"
    assert str(Float("-inf")) == "-inf"
    assert str(Float("infinity")) == "inf"
    assert str(Float("+infinity")) == "inf"
    assert str(Float("-infinity")) == "-inf"

    with pytest.raises(ValueError):
        Float("info")
    with pytest.raises(ValueError):
        Float("+info")
    with pytest.raises(ValueError):
        Float("-info")
    with pytest.raises(ValueError):
        Float("in")
    with pytest.raises(ValueError):
        Float("+in")
    with pytest.raises(ValueError):
        Float("-in")
    with pytest.raises(ValueError):
        Float("infinit")
    with pytest.raises(ValueError):
        Float("+Infin")
    with pytest.raises(ValueError):
        Float("-INFI")
    with pytest.raises(ValueError):
        Float("infinitys")

    with pytest.raises(ValueError):
        Float("++Inf")
    with pytest.raises(ValueError):
        Float("-+inf")
    with pytest.raises(ValueError):
        Float("+-infinity")
    with pytest.raises(ValueError):
        Float("--Infinity")


def test_inf_as_str():
    assert repr(1e300 * 1e300) == "inf"
    assert repr(-1e300 * 1e300) == "-inf"

    assert str(1e300 * 1e300) == "inf"
    assert str(-1e300 * 1e300) == "-inf"


def test_nan_from_str():
    assert isnan(Float("nan"))
    assert isnan(Float("+nan"))
    assert isnan(Float("-nan"))

    assert repr(Float("nan")) == "nan"
    assert repr(Float("+nan")) == "nan"
    assert repr(Float("-nan")) == "nan"

    assert repr(Float("NAN")) == "nan"
    assert repr(Float("+NAn")) == "nan"
    assert repr(Float("-NaN")) == "nan"

    assert str(Float("nan")) == "nan"
    assert str(Float("+nan")) == "nan"
    assert str(Float("-nan")) == "nan"

    with pytest.raises(ValueError):
        Float("nana")
    with pytest.raises(ValueError):
        Float("+nana")
    with pytest.raises(ValueError):
        Float("-nana")
    with pytest.raises(ValueError):
        Float("na")
    with pytest.raises(ValueError):
        Float("+na")
    with pytest.raises(ValueError):
        Float("-na")

    with pytest.raises(ValueError):
        Float("++nan")
    with pytest.raises(ValueError):
        Float("-+NAN")
    with pytest.raises(ValueError):
        Float("+-NaN")
    with pytest.raises(ValueError):
        Float("--nAn")


def test_nan_as_str():
    assert repr(1e300 * 1e300 * 0) == "nan"
    assert repr(-1e300 * 1e300 * 0) == "nan"

    assert str(1e300 * 1e300 * 0) == "nan"
    assert str(-1e300 * 1e300 * 0) == "nan"


def test_inf_signs():
    assert copysign(1.0, Float("inf")) == 1.0
    assert copysign(1.0, Float("-inf")) == -1.0


def test_nan_signs():
    # When using the dtoa.c code, the sign of Float('nan') should
    # be predictable.
    assert copysign(1.0, Float("nan")) == 1.0
    assert copysign(1.0, Float("-nan")) == -1.0


fromHex = Float.fromhex
toHex = Float.hex

MAX = fromHex("0x.fffffffffffff8p+1024")  # max normal
MIN = fromHex("0x1p-1022")  # min normal
TINY = fromHex("0x0.0000000000001p-1022")  # min subnormal
EPS = fromHex("0x0.0000000000001p0")  # diff between 1.0 and next Float up


def identical(x, y):
    # check that Floats x and y are identical, or that both
    # are NaNs
    if isnan(x) or isnan(y):
        if isnan(x) == isnan(y):
            return
    elif x == y and (x != 0.0 or copysign(1.0, x) == copysign(1.0, y)):
        return
    pytest.fail(f"{x!r} not identical to {y!r}")


def test_ends():
    identical(MIN, ldexp(1.0, -1022))
    identical(TINY, ldexp(1.0, -1074))
    identical(EPS, ldexp(1.0, -52))
    identical(MAX, 2.0 * (ldexp(1.0, 1023) - ldexp(1.0, 970)))


def test_invalid_inputs():
    invalid_inputs = [
        "infi",  # misspelt infinities and nans
        "-Infinit",
        "++inf",
        "-+Inf",
        "--nan",
        "+-NaN",
        "snan",
        "NaNs",
        "nna",
        "an",
        "nf",
        "nfinity",
        "inity",
        "iinity",
        "0xnan",
        "",
        " ",
        "x1.0p0",
        "0xX1.0p0",
        "+ 0x1.0p0",  # internal whitespace
        "- 0x1.0p0",
        "0 x1.0p0",
        "0x 1.0p0",
        "0x1 2.0p0",
        "+0x1 .0p0",
        "0x1. 0p0",
        "-0x1.0 1p0",
        "-0x1.0 p0",
        "+0x1.0p +0",
        "0x1.0p -0",
        "0x1.0p 0",
        "+0x1.0p+ 0",
        "-0x1.0p- 0",
        "++0x1.0p-0",  # double signs
        "--0x1.0p0",
        "+-0x1.0p+0",
        "-+0x1.0p0",
        "0x1.0p++0",
        "+0x1.0p+-0",
        "-0x1.0p-+0",
        "0x1.0p--0",
        "0x1.0.p0",
        "0x.p0",  # no hex digits before or after point
        "0x1,p0",  # wrong decimal point character
        "0x1pa",
        "0x1p\uff10",  # fullwidth Unicode digits
        "\uff10x1p0",
        "0x\uff11p0",
        "0x1.\uff10p0",
        "0x1p0 \n 0x2p0",
        "0x1p0\0 0x1p0",  # embedded null byte is not end of string
    ]
    for x in invalid_inputs:
        try:
            result = fromHex(x)
        except ValueError:
            pass
        else:
            pytest.fail(
                "Expected Float.fromhex(%r) to raise ValueError; "
                "got %r instead" % (x, result)
            )



# stdlib
import json
import os
import sys
from typing import Any
from typing import Generator
from typing import Optional

# third party
import requests

# syft relative
from ...core.common.environment import is_jupyter
from ...core.node.common.client import Client
from ...core.node.domain.domain import Domain
from ...logger import error
from ...logger import info
from ...logger import traceback_and_raise
from .bcolors import bcolors
from .exchange_ids import DuetCredentialExchanger
from .exchange_ids import OpenGridTokenFileExchanger
from .exchange_ids import OpenGridTokenManualInputExchanger
from .om_signaling_client import register
from .ui import LOGO_URL
from .webrtc_duet import Duet as WebRTCDuet  # noqa: F811

if is_jupyter:
    # third party
    from IPython.core.display import Image
    from IPython.core.display import display


ADDR_REPOSITORY = (
    "https://raw.githubusercontent.com/OpenMined/OpenGridNodes/master/network_address"
)


def generate_donation_msg(name: str) -> str:
    donate_url = "https://github.com/sponsors/OpenMined"
    donate_msg = (
        f"\n    > ❤️ {bcolors.FAIL}Love{bcolors.ENDC} {bcolors.OKGREEN}{name}{bcolors.ENDC}? "
        + f"{bcolors.WARNING}Please{bcolors.ENDC} {bcolors.OKBLUE}consider{bcolors.ENDC} "
        + f"{bcolors.HEADER}supporting{bcolors.ENDC} {bcolors.FAIL}our{bcolors.ENDC} "
        + f"{bcolors.WARNING}community!{bcolors.ENDC}"
        + f"\n    > {donate_url}"
    )
    return donate_msg


DUET_DONATE_MSG = generate_donation_msg(name="Duet")


def get_available_network() -> str:
    network_addr = json.loads(requests.get(ADDR_REPOSITORY).content)
    for addr in network_addr:
        try:
            requests.get(addr + "/metadata")
            return addr
        except Exception as e:
            error(f"Failed request addr: {e}")
            continue
    traceback_and_raise(Exception("Couldn't find any available network."))


def begin_duet_logger(my_domain: Domain) -> None:
    # stdlib
    from contextlib import contextmanager
    import threading
    import time

    # we need a lock, so that other threads don't snatch control
    # while we have set a temporary parent
    stdout_lock = threading.Lock()

    @contextmanager
    def set_stdout_parent(parent: Any) -> Generator:
        """a context manager for setting a particular parent for sys.stdout
        the parent determines the destination cell of output
        """
        save_parent = sys.stdout.parent_header  # type: ignore
        with stdout_lock:
            sys.stdout.parent_header = parent  # type: ignore
            try:
                yield
            finally:
                # the flush is important, because that's when the parent_header actually has its effect
                sys.stdout.flush()
                sys.stdout.parent_header = save_parent  # type: ignore

    class counterThread(threading.Thread):
        def run(self) -> None:
            # record the parent when the thread starts
            thread_parent = sys.stdout.parent_header  # type: ignore
            iterator = 0
            while True:
                time.sleep(0.1)
                # then ensure that the parent is the same as when the thread started
                # every time we print
                with set_stdout_parent(thread_parent):

                    n_objects = len(my_domain.store)
                    n_requests = len(my_domain.requests)
                    n_messages = my_domain.message_counter
                    n_request_handlers = len(my_domain.request_handlers)

                    blink_on = (int(iterator / 5) % 2) == 0

                    if blink_on and n_requests > 0:
                        left_blink = bcolors.BOLD + ">" + bcolors.ENDC
                        right_blink = bcolors.BOLD + "<" + bcolors.ENDC
                        left_color = bcolors.FAIL
                        right_color = bcolors.ENDC
                    else:
                        left_blink = " "
                        right_blink = " "
                        left_color = ""
                        right_color = ""

                    if blink_on:
                        star = "*"
                    else:
                        star = "-"

                    out = (
                        "♫♫♫ > DUET LIVE STATUS  "
                        + star
                        + "  Objects: "
                        + str(n_objects)
                        + "  "
                        + left_color
                        + "Requests:"
                        + right_color
                        + left_blink
                        + str(n_requests)
                        + right_blink
                        + "  Messages: "
                        + str(n_messages)
                        + "  Request Handlers: "
                        + str(n_request_handlers)
                    )
                    out += "                                "
                    # STOP changing this to logging, this happens every fraction of a
                    # second to update the jupyter display, logging this creates
                    # unnecessary noise, in addition the end= parameter broke logging
                    print("\r" + out, end="\r")  # DO NOT change to log
                iterator += 1

    if hasattr(sys.stdout, "parent_header"):
        counterThread().start()


def duet(
    target_id: Optional[str] = None,
    logging: bool = True,
    network_url: str = "",
    loopback: bool = False,
    db_path: Optional[str] = None,
) -> Client:
    if target_id is not None:
        return join_duet(
            target_id=target_id, loopback=loopback, network_url=network_url
        )
    else:
        return launch_duet(
            logging=logging, network_url=network_url, loopback=loopback, db_path=db_path
        )


def launch_duet(
    logging: bool = True,
    network_url: str = "",
    loopback: bool = False,
    credential_exchanger: DuetCredentialExchanger = OpenGridTokenManualInputExchanger(),
    db_path: Optional[str] = None,
) -> Client:
    if os.path.isfile(LOGO_URL) and is_jupyter:
        display(
            Image(
                LOGO_URL,
                width=400,
                unconfined=True,
            )
        )
    info("🎤  🎸  ♪♪♪ Starting Duet ♫♫♫  🎻  🎹\n", print=True)
    info(
        "♫♫♫ >\033[93m"
        + " DISCLAIMER"
        + "\033[0m"
        + ": "
        + "\033[1m"
        + "Duet is an experimental feature currently in beta.\n"
        + "♫♫♫ > Use at your own risk.\n"
        + "\033[0m",
        print=True,
    )

    info(bcolors.BOLD + DUET_DONATE_MSG + bcolors.BOLD + "\n", print=True)

    if not network_url:
        network_url = get_available_network()
    info("♫♫♫ > Punching through firewall to OpenGrid Network Node at:", print=True)
    info("♫♫♫ > " + str(network_url), print=True)
    info("♫♫♫ >", print=True)
    info("♫♫♫ > ...waiting for response from OpenGrid Network... ", print=True)

    signaling_client = register(url=network_url)

    info("♫♫♫ > " + bcolors.OKGREEN + "DONE!" + bcolors.ENDC, print=True)

    my_domain = Domain(name="Launcher", db_path=db_path)

    if loopback:
        credential_exchanger = OpenGridTokenFileExchanger()
    target_id = credential_exchanger.run(credential=signaling_client.duet_id)

    info("♫♫♫ > Connecting...", print=True)

    _ = WebRTCDuet(
        node=my_domain,
        target_id=target_id,
        signaling_client=signaling_client,
        offer=True,
    )
    info(print=True)
    info("♫♫♫ > " + bcolors.OKGREEN + "CONNECTED!" + bcolors.ENDC, print=True)
    #     return duet, my_domain.get_root_client()
    out_duet: Client = my_domain.get_root_client()

    if logging:
        begin_duet_logger(my_domain=my_domain)
    info(print=True)

    return out_duet


def join_duet(
    target_id: str = "",
    network_url: str = "",
    loopback: bool = False,
    credential_exchanger: DuetCredentialExchanger = OpenGridTokenManualInputExchanger(),
) -> WebRTCDuet:
    if os.path.isfile(LOGO_URL) and is_jupyter:
        display(
            Image(
                LOGO_URL,
                width=400,
                unconfined=True,
            )
        )
    info("🎤  🎸  ♪♪♪ Joining Duet ♫♫♫  🎻  🎹\n", print=True)
    info(
        "♫♫♫ >\033[93m"
        + " DISCLAIMER"
        + "\033[0m"
        + ": "
        + "\033[1m"
        + "Duet is an experimental feature currently in beta.\n"
        + "♫♫♫ > Use at your own risk.\n"
        + "\033[0m",
        print=True,
    )

    info(bcolors.BOLD + DUET_DONATE_MSG + bcolors.BOLD + "\n", print=True)

    if not network_url:
        network_url = get_available_network()
    info("♫♫♫ > Punching through firewall to OpenGrid Network Node at:", print=True)
    info("♫♫♫ > " + str(network_url), print=True)
    info("♫♫♫ >", print=True)
    info("♫♫♫ > ...waiting for response from OpenGrid Network... ", print=True)

    signaling_client = register(url=network_url)

    info("♫♫♫ > " + bcolors.OKGREEN + "DONE!" + bcolors.ENDC, print=True)

    my_domain = Domain(name="Joiner")

    if loopback:
        credential_exchanger = OpenGridTokenFileExchanger()
    else:
        # we have target_id so we set it using set_responder_id
        credential_exchanger.set_responder_id(target_id)

    target_id = credential_exchanger.set_role(join=True).run(
        credential=signaling_client.duet_id
    )

    duet = WebRTCDuet(
        node=my_domain,
        target_id=target_id,
        signaling_client=signaling_client,
        offer=False,
    )
    info(print=True)
    info("♫♫♫ > " + bcolors.OKGREEN + "CONNECTED!" + bcolors.ENDC, print=True)
    # begin_duet_client_logger(duet.node)

    return duet



# stdlib
from typing import Optional
from typing import Union

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft relative
from ....logger import error
from ...common.message import SignedMessage
from ...common.message import SyftMessage
from ...common.uid import UID
from ...io.location import Location
from ...io.location import SpecificLocation
from ..common.node import Node
from ..domain.client import DomainClient
from ..domain.domain import Domain
from .client import NetworkClient


class Network(Node):

    network: SpecificLocation

    child_type = Domain
    client_type = NetworkClient
    child_type_client_type = DomainClient

    def __init__(
        self,
        name: Optional[str],
        network: SpecificLocation = SpecificLocation(),
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
    ):
        super().__init__(
            name=name,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
        )

        # specific location with name
        self.network = SpecificLocation(name=self.name)

        self._register_services()
        self.post_init()

    @property
    def icon(self) -> str:
        return "🔗"

    @property
    def id(self) -> UID:
        return self.network.id

    def message_is_for_me(self, msg: Union[SyftMessage, SignedMessage]) -> bool:
        # this needs to be defensive by checking network_id NOT network.id or it breaks
        try:
            return msg.address.network_id == self.id and msg.address.domain is None
        except Exception as e:
            error(f"Error checking if {msg.pprint} is for me on {self.pprint}. {e}")
            return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/returntypes.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.lib.torch import tensor_pb2 as proto_dot_lib_dot_torch_dot_tensor__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/torch/returntypes.proto",
    package="syft.lib.torch",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n!proto/lib/torch/returntypes.proto\x12\x0esyft.lib.torch\x1a\x1cproto/lib/torch/tensor.proto"K\n\x0bReturnTypes\x12\x10\n\x08obj_type\x18\x01 \x01(\t\x12*\n\x06values\x18\x02 \x03(\x0b\x32\x1a.syft.lib.torch.TensorDatab\x06proto3',
    dependencies=[
        proto_dot_lib_dot_torch_dot_tensor__pb2.DESCRIPTOR,
    ],
)


_RETURNTYPES = _descriptor.Descriptor(
    name="ReturnTypes",
    full_name="syft.lib.torch.ReturnTypes",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="obj_type",
            full_name="syft.lib.torch.ReturnTypes.obj_type",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="values",
            full_name="syft.lib.torch.ReturnTypes.values",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=83,
    serialized_end=158,
)

_RETURNTYPES.fields_by_name[
    "values"
].message_type = proto_dot_lib_dot_torch_dot_tensor__pb2._TENSORDATA
DESCRIPTOR.message_types_by_name["ReturnTypes"] = _RETURNTYPES
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ReturnTypes = _reflection.GeneratedProtocolMessageType(
    "ReturnTypes",
    (_message.Message,),
    {
        "DESCRIPTOR": _RETURNTYPES,
        "__module__": "proto.lib.torch.returntypes_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.ReturnTypes)
    },
)
_sym_db.RegisterMessage(ReturnTypes)


# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Any
from typing import Dict

# syft absolute
import syft as sy
from syft import serialize
from syft.core.io.address import Address
from syft.grid.messages.role_messages import CreateRoleMessage
from syft.grid.messages.role_messages import CreateRoleResponse
from syft.grid.messages.role_messages import DeleteRoleMessage
from syft.grid.messages.role_messages import DeleteRoleResponse
from syft.grid.messages.role_messages import GetRoleMessage
from syft.grid.messages.role_messages import GetRoleResponse
from syft.grid.messages.role_messages import GetRolesMessage
from syft.grid.messages.role_messages import GetRolesResponse
from syft.grid.messages.role_messages import UpdateRoleMessage
from syft.grid.messages.role_messages import UpdateRoleResponse


def test_create_role_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {
        "name": "Role Sample",
        "can_triage_results": True,
        "can_edit_settings": False,
        "can_create_users": False,
        "can_edit_roles": False,
        "can_manage_infrastructure": True,
    }
    msg = CreateRoleMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_create_role_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Role created succesfully!"}
    msg = CreateRoleResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_role_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {"role_id": "f2a6as5d16fasd"}
    msg = DeleteRoleMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_role_response_serde() -> None:
    target = Address(name="Alice")

    content = {"msg": "Role has been deleted!"}
    msg = DeleteRoleResponse(
        status_code=200,
        address=target,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_role_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {
        "role_id": "9a4f9dasd6",
        "name": "Role Sample",
        "can_triage_results": True,
        "can_edit_settings": False,
        "can_create_users": False,
        "can_edit_roles": False,
        "can_manage_infrastructure": True,
    }
    msg = UpdateRoleMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_role_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Role has been updated successfully!"}
    msg = UpdateRoleResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_role_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {"request_id": "eqw9e4a5d846"}
    msg = GetRoleMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_role_response_serde() -> None:
    target = Address(name="Alice")

    content = {
        "name": "Role Sample",
        "can_triage_results": True,
        "can_edit_settings": False,
        "can_create_users": False,
        "can_edit_roles": False,
        "can_manage_infrastructure": True,
    }

    msg = GetRoleResponse(
        address=target,
        status_code=200,
        content=content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_roles_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content: Dict[Any, Any] = {}
    msg = GetRolesMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_roles_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {
        "workers": {
            "626sadaf631": {
                "name": "Role Sample",
                "can_triage_results": True,
                "can_edit_settings": False,
                "can_create_users": False,
                "can_edit_roles": False,
                "can_manage_infrastructure": True,
            },
            "a84ew64wq6e": {
                "name": "Test Sample",
                "can_triage_results": False,
                "can_edit_settings": True,
                "can_create_users": False,
                "can_edit_roles": False,
                "can_manage_infrastructure": False,
            },
        }
    }

    msg = GetRolesResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2



# stdlib
import atexit
from multiprocessing import Process
from multiprocessing import set_start_method
import socket
import sys
import time
from typing import Any
from typing import Generator
from typing import List

# third party
import pytest

# syft absolute
import syft as sy

# syft relative
from ...grid.duet.process_test import SyftTestProcess
from ...grid.duet.signaling_server_test import run

ts = pytest.importorskip("tenseal")
sy.load("tenseal")

set_start_method("spawn", force=True)
PORT = 21000


def chunks(lst: List[Any], n: int) -> Generator[Any, Any, Any]:
    """Yield successive n-sized chunks from lst.

    Args:
        lst: list of items to chunk
        n: number of items to include in each chunk

    Yields:
        single chunk of n items
    """
    for i in range(0, len(lst), n):
        yield lst[i : i + n]  # noqa: E203


def do(ct_size: int, batch_size: int) -> None:
    # third party
    import numpy as np
    import tenseal as ts

    # syft absolute
    import syft as sy

    sy.load("tenseal")
    sy.logger.add(sys.stderr, "ERROR")

    duet = sy.launch_duet(loopback=True, network_url=f"http://127.0.0.1:{PORT}/")
    duet.requests.add_handler(action="accept")

    context = ts.context(
        ts.SCHEME_TYPE.CKKS, 8192, coeff_mod_bit_sizes=[60, 40, 40, 60], n_threads=1
    )
    context.global_scale = pow(2, 40)

    data = np.random.uniform(-10, 10, 100)
    enc = []
    for i in range(ct_size):
        enc.append(ts.ckks_vector(context, data))

    start = time.time()
    _ = context.send(duet, pointable=True)
    for chunk in chunks(enc, batch_size):
        _ = sy.lib.python.List(chunk).send(duet, pointable=True)
    sys.stderr.write(
        f"[{ct_size}][{batch_size}] DO sending took {time.time() - start} sec\n"
    )

    sy.core.common.event_loop.loop.run_forever()


def ds(ct_size: int, batch_size: int) -> None:
    # syft absolute
    import syft as sy

    sy.load("tenseal")
    sy.logger.add(sys.stderr, "ERROR")

    duet = sy.join_duet(loopback=True, network_url=f"http://127.0.0.1:{PORT}/")

    time.sleep(10)

    cnt = int(ct_size / batch_size)

    start = time.time()
    ctx = duet.store[0].get(request_block=True, delete_obj=False)
    for idx in range(1, cnt + 1):
        data = duet.store[idx].get(request_block=True, delete_obj=False)

        for tensor in data:
            tensor.link_context(ctx)

        assert len(data) == batch_size, len(data)

    sys.stderr.write(
        f"[{ct_size}][{batch_size}] DS get took {time.time() - start} sec\n"
    )


@pytest.fixture(scope="module")
def signaling_server() -> Process:
    print(f"creating signaling server on port {PORT}")
    grid_proc = Process(target=run, args=(PORT,))
    grid_proc.start()

    def grid_cleanup() -> None:
        print("stop signaling server")
        grid_proc.terminate()
        grid_proc.join()

    atexit.register(grid_cleanup)

    return grid_proc


@pytest.mark.vendor(lib="tenseal")
def test_tenseal_duet_ciphertext_size(signaling_server: Process) -> None:
    time.sleep(3)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        assert s.connect_ex(("localhost", PORT)) == 0

    for ct_size in [10, 20]:
        for batch_size in [1, 10, ct_size]:
            start = time.time()

            do_proc = SyftTestProcess(target=do, args=(ct_size, batch_size))
            do_proc.start()

            ds_proc = SyftTestProcess(target=ds, args=(ct_size, batch_size))
            ds_proc.start()

            ds_proc.join(120)

            do_proc.terminate()

            if do_proc.exception:
                exception, tb = do_proc.exception
                raise Exception(tb) from exception

            if ds_proc.exception:
                exception, tb = ds_proc.exception
                raise Exception(tb) from exception

            if ds_proc.is_alive():
                ds_proc.terminate()
                raise Exception(f"ds_proc is hanged for {ct_size}")

            print(
                f"test {ct_size} batch_size {batch_size} passed in {time.time() - start} seconds"
            )



# stdlib
from typing import Any
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from ... import deserialize
from ... import serialize
from ...core.common import UID
from ...core.common.serde.serializable import bind_protobuf
from ...proto.lib.python.tuple_pb2 import Tuple as Tuple_PB
from .iterator import Iterator
from .primitive_factory import PrimitiveFactory
from .primitive_factory import isprimitive
from .primitive_interface import PyPrimitive
from .slice import Slice
from .types import SyPrimitiveRet
from .util import downcast
from .util import upcast


class TupleIterator(Iterator):
    pass


@bind_protobuf
class Tuple(tuple, PyPrimitive):
    def __init__(self, *args: Any):
        self._id = UID()

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def upcast(self) -> tuple:
        # recursively upcast
        return tuple(upcast(v) for v in self)

    def __new__(cls, *args: Any) -> "Tuple":
        return super(Tuple, cls).__new__(Tuple, *args)

    def __add__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__add__(other))

    def __contains__(self, item: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__contains__(item))

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__eq__(other))

    def __hash__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__hash__())

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__ne__(other))

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__ge__(other))

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__gt__(other))

    def __le__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__le__(other))

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__lt__(other))

    def __mul__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__mul__(other))

    def __rmul__(self, other: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__rmul__(other))

    def __len__(self) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().__len__())

    def __getitem__(self, key: Union[int, slice, Slice]) -> Any:
        if isinstance(key, Slice):
            key = key.upcast()
        value = super().__getitem__(key)
        if isprimitive(value=value):
            return PrimitiveFactory.generate_primitive(value=value)
        else:
            # we can have torch.Tensor and other types
            return value

    def count(self, __value: Any) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().count(__value))

    def index(
        self, __value: Any, __start: Any = ..., __stop: Any = ...
    ) -> SyPrimitiveRet:
        return PrimitiveFactory.generate_primitive(value=super().index(__value, __stop))

    def __iter__(self, max_len: Optional[int] = None) -> TupleIterator:
        return TupleIterator(self, max_len=max_len)

    def _object2proto(self) -> Tuple_PB:
        id_ = serialize(obj=self.id)
        downcasted = [downcast(value=element) for element in self]
        data = [serialize(obj=element, to_bytes=True) for element in downcasted]
        return Tuple_PB(id=id_, data=data)

    @staticmethod
    def _proto2object(proto: Tuple_PB) -> "Tuple":
        id_: UID = deserialize(blob=proto.id)
        value = [
            upcast(deserialize(blob=element, from_bytes=True)) for element in proto.data
        ]
        new_list = Tuple(value)
        new_list._id = id_
        return new_list

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Tuple_PB



# stdlib
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# syft relative
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ...core.remote_dataloader import RemoteDataLoader
from ...core.remote_dataloader import RemoteDataset


def create_remote_dataloader_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules = [
        "syft",
        "syft.core",
        "syft.core.remote_dataloader",
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        (
            "syft.core.remote_dataloader.RemoteDataset",
            "syft.core.remote_dataloader.RemoteDataset",
            RemoteDataset,
        ),
        (
            "syft.core.remote_dataloader.RemoteDataLoader",
            "syft.core.remote_dataloader.RemoteDataLoader",
            RemoteDataLoader,
        ),
    ]

    methods: TypeList[TypeTuple[str, str]] = [
        (
            "syft.core.remote_dataloader.RemoteDataset.load_dataset",
            "syft.lib.python._SyNone",
        ),
        ("syft.core.remote_dataloader.RemoteDataset.__len__", "syft.lib.python.Int"),
        ("syft.core.remote_dataloader.RemoteDataset.__getitem__", "torch.Tensor"),
        ("syft.core.remote_dataloader.RemoteDataLoader.__len__", "syft.lib.python.Int"),
        (
            "syft.core.remote_dataloader.RemoteDataLoader.load_dataset",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.core.remote_dataloader.RemoteDataLoader.create_dataloader",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.core.remote_dataloader.RemoteDataLoader.__iter__",
            "syft.lib.python.Iterator",
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/group_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/grid/messages/group_messages.proto",
    package="syft.grid.messages",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n(proto/grid/messages/group_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x9d\x01\n\x12\x43reateGroupMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8a\x01\n\x13\x43reateGroupResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9a\x01\n\x0fGetGroupMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x87\x01\n\x10GetGroupResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9b\x01\n\x10GetGroupsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x88\x01\n\x11GetGroupsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9d\x01\n\x12\x44\x65leteGroupMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8a\x01\n\x13\x44\x65leteGroupResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9d\x01\n\x12UpdateGroupMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8a\x01\n\x13UpdateGroupResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
    ],
)


_CREATEGROUPMESSAGE = _descriptor.Descriptor(
    name="CreateGroupMessage",
    full_name="syft.grid.messages.CreateGroupMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.CreateGroupMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.CreateGroupMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.CreateGroupMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.CreateGroupMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=133,
    serialized_end=290,
)


_CREATEGROUPRESPONSE = _descriptor.Descriptor(
    name="CreateGroupResponse",
    full_name="syft.grid.messages.CreateGroupResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.CreateGroupResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.CreateGroupResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.CreateGroupResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.CreateGroupResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=293,
    serialized_end=431,
)


_GETGROUPMESSAGE = _descriptor.Descriptor(
    name="GetGroupMessage",
    full_name="syft.grid.messages.GetGroupMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetGroupMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetGroupMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetGroupMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.GetGroupMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=434,
    serialized_end=588,
)


_GETGROUPRESPONSE = _descriptor.Descriptor(
    name="GetGroupResponse",
    full_name="syft.grid.messages.GetGroupResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetGroupResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.GetGroupResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetGroupResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetGroupResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=591,
    serialized_end=726,
)


_GETGROUPSMESSAGE = _descriptor.Descriptor(
    name="GetGroupsMessage",
    full_name="syft.grid.messages.GetGroupsMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetGroupsMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetGroupsMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetGroupsMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.GetGroupsMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=729,
    serialized_end=884,
)


_GETGROUPSRESPONSE = _descriptor.Descriptor(
    name="GetGroupsResponse",
    full_name="syft.grid.messages.GetGroupsResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.GetGroupsResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.GetGroupsResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.GetGroupsResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.GetGroupsResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=887,
    serialized_end=1023,
)


_DELETEGROUPMESSAGE = _descriptor.Descriptor(
    name="DeleteGroupMessage",
    full_name="syft.grid.messages.DeleteGroupMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.DeleteGroupMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.DeleteGroupMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.DeleteGroupMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.DeleteGroupMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1026,
    serialized_end=1183,
)


_DELETEGROUPRESPONSE = _descriptor.Descriptor(
    name="DeleteGroupResponse",
    full_name="syft.grid.messages.DeleteGroupResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.DeleteGroupResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.DeleteGroupResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.DeleteGroupResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.DeleteGroupResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1186,
    serialized_end=1324,
)


_UPDATEGROUPMESSAGE = _descriptor.Descriptor(
    name="UpdateGroupMessage",
    full_name="syft.grid.messages.UpdateGroupMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.UpdateGroupMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.UpdateGroupMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.UpdateGroupMessage.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.messages.UpdateGroupMessage.reply_to",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1327,
    serialized_end=1484,
)


_UPDATEGROUPRESPONSE = _descriptor.Descriptor(
    name="UpdateGroupResponse",
    full_name="syft.grid.messages.UpdateGroupResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.messages.UpdateGroupResponse.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="status_code",
            full_name="syft.grid.messages.UpdateGroupResponse.status_code",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="syft.grid.messages.UpdateGroupResponse.content",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.messages.UpdateGroupResponse.address",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1487,
    serialized_end=1625,
)

_CREATEGROUPMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_CREATEGROUPMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_CREATEGROUPMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_CREATEGROUPRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_CREATEGROUPRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETGROUPMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETGROUPMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETGROUPMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETGROUPRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETGROUPRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETGROUPSMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETGROUPSMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETGROUPSMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_GETGROUPSRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_GETGROUPSRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_DELETEGROUPMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_DELETEGROUPMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_DELETEGROUPMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_DELETEGROUPRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_DELETEGROUPRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_UPDATEGROUPMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_UPDATEGROUPMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_UPDATEGROUPMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_UPDATEGROUPRESPONSE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_UPDATEGROUPRESPONSE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
DESCRIPTOR.message_types_by_name["CreateGroupMessage"] = _CREATEGROUPMESSAGE
DESCRIPTOR.message_types_by_name["CreateGroupResponse"] = _CREATEGROUPRESPONSE
DESCRIPTOR.message_types_by_name["GetGroupMessage"] = _GETGROUPMESSAGE
DESCRIPTOR.message_types_by_name["GetGroupResponse"] = _GETGROUPRESPONSE
DESCRIPTOR.message_types_by_name["GetGroupsMessage"] = _GETGROUPSMESSAGE
DESCRIPTOR.message_types_by_name["GetGroupsResponse"] = _GETGROUPSRESPONSE
DESCRIPTOR.message_types_by_name["DeleteGroupMessage"] = _DELETEGROUPMESSAGE
DESCRIPTOR.message_types_by_name["DeleteGroupResponse"] = _DELETEGROUPRESPONSE
DESCRIPTOR.message_types_by_name["UpdateGroupMessage"] = _UPDATEGROUPMESSAGE
DESCRIPTOR.message_types_by_name["UpdateGroupResponse"] = _UPDATEGROUPRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CreateGroupMessage = _reflection.GeneratedProtocolMessageType(
    "CreateGroupMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEGROUPMESSAGE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateGroupMessage)
    },
)
_sym_db.RegisterMessage(CreateGroupMessage)

CreateGroupResponse = _reflection.GeneratedProtocolMessageType(
    "CreateGroupResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEGROUPRESPONSE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateGroupResponse)
    },
)
_sym_db.RegisterMessage(CreateGroupResponse)

GetGroupMessage = _reflection.GeneratedProtocolMessageType(
    "GetGroupMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETGROUPMESSAGE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetGroupMessage)
    },
)
_sym_db.RegisterMessage(GetGroupMessage)

GetGroupResponse = _reflection.GeneratedProtocolMessageType(
    "GetGroupResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETGROUPRESPONSE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetGroupResponse)
    },
)
_sym_db.RegisterMessage(GetGroupResponse)

GetGroupsMessage = _reflection.GeneratedProtocolMessageType(
    "GetGroupsMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETGROUPSMESSAGE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetGroupsMessage)
    },
)
_sym_db.RegisterMessage(GetGroupsMessage)

GetGroupsResponse = _reflection.GeneratedProtocolMessageType(
    "GetGroupsResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETGROUPSRESPONSE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetGroupsResponse)
    },
)
_sym_db.RegisterMessage(GetGroupsResponse)

DeleteGroupMessage = _reflection.GeneratedProtocolMessageType(
    "DeleteGroupMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEGROUPMESSAGE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteGroupMessage)
    },
)
_sym_db.RegisterMessage(DeleteGroupMessage)

DeleteGroupResponse = _reflection.GeneratedProtocolMessageType(
    "DeleteGroupResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEGROUPRESPONSE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteGroupResponse)
    },
)
_sym_db.RegisterMessage(DeleteGroupResponse)

UpdateGroupMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateGroupMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEGROUPMESSAGE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateGroupMessage)
    },
)
_sym_db.RegisterMessage(UpdateGroupMessage)

UpdateGroupResponse = _reflection.GeneratedProtocolMessageType(
    "UpdateGroupResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEGROUPRESPONSE,
        "__module__": "proto.grid.messages.group_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateGroupResponse)
    },
)
_sym_db.RegisterMessage(UpdateGroupResponse)


# @@protoc_insertion_point(module_scope)



# third party
import pytest

# syft absolute
import syft as sy


@pytest.mark.vendor(lib="opacus")
def test_remote_engine_simple(root_client: sy.VirtualMachineClient) -> None:
    sy.load("opacus")

    remote_opacus = root_client.opacus
    remote_torch = root_client.torch

    model_ptr = remote_torch.nn.Linear(1, 1)
    batch_size = 16
    sample_size = 16
    noise_multiplier = 1.0
    max_grad_norm = 1.0

    privacy_engine_ptr = remote_opacus.privacy_engine.PrivacyEngine(
        model_ptr,
        batch_size=batch_size,
        sample_size=sample_size,
        noise_multiplier=noise_multiplier,
        max_grad_norm=max_grad_norm,
    )

    assert privacy_engine_ptr.__name__ == "PrivacyEnginePointer"
